#ifndef __menu_h__
#define __menu_h__

#include <string>
#include <vector>
#include "actionhandler.h"
#include "surface.h"
#include "configfile.h"

using namespace std;

/*

	Base class for all the Menu objects. If you're too lazy to read any source, read at least this one :)

*/

namespace hm
{

	/**
	*	The menu class represents a node in the menu tree. Visually it usually has two forms: an item in a list 
	*	and a window the user can interact with. (This is not always the case though, as with the CheckboxMenu)
	*	
	*	A menu system can be constructed by creating a single root Menu object, and adding other Menu objects to it.
	*	Various subclasses of the Menu class exist, like the CheckboxMenu class, SelectMenu class and ActionMenu class.
	*	
	*/

	class Menu
   {

	public:

		/** 
			constructor
			\param title the title of the menu, it will usually show in the list and on the window header
			\param surface a surface to draw on. The root node should always have a non-NULL Surface object
			specified, child nodes will inherit their parent's Surface when appendChild is called.
		*/
		Menu(string title,Surface *surface = NULL);

		/** destructor */
		virtual ~Menu();
		
		/** draws the menu as a list entry 
			\param surface a Surface object to draw to. The reason the Menu's own surface isn't automatically used
					is because of dual screen support (DS port anyone?)
			\param x horizontal position
			\param y vertical position
			\param w width
			\param h height
			\param selected indicates wether the cursor is over this listitem
		*/
		virtual void drawListItem(Surface *surface,int x,int y,int w,int h,bool selected);

		/** \returns <int> the height of the listitem representation of this Menu */
		virtual int getListItemHeight();

		/** called when this Menu is selected in the parent menu. */
		virtual int enter();

		/** change the window size 
			\param w width
			\param h height
		*/
		void setWindowSize(int w,int h);

		/** change the window position 
			\param x horizontal position
			\param y vertical position
		*/
		void setWindowPosition(int x,int y);

		/** append a Menu item to this menu and takes it's ownership
			\param child the Menu object to be added
		*/
		void appendChild(Menu* child);

		/** set the icon for this item 
			\param icon an Image object to act as a icon. The icon will be shown in front of the title (unless drawListItem
				is overridden to act differently)
		*/
		void setIcon(Image* icon);

		/** returns the icon as an Image object
		*/
		Image* getIcon();

		/** set the secondary icon for this item 
			\param icon an Image object to act as a secondary icon. Is used by RadMenu now, but could be used in other
			menu's as well
		*/
		void setSecondaryIcon(Image* icon);

		/** returns the secondary icon as an Image object
		*/
		Image* getSecondaryIcon();

		/** sets the name of this menu. If the name is set to something else than the emtpy string it will automatically
			populate the ConfigFile object (see addValues).
			\param name the new name
		*/
		void setName(string name);

		/** sets the value of this menu. The value of a Menu usually represents it's state. A CheckboxMenu can either
			have the valye "true" or "false", whereas a SelectMenu's value can take the form of any of it's children's 
			values.
			\param value the new value
		*/
		void setValue(string value);

		/** sets the title of this menu. The title is shown in the list and window representations.
			\param text the new title
		*/
		void setTitle(string text);

		/** \returns <string> the name of this Menu */
		string getName();

		/** \returns <string> the value of this menu. The value of a Menu usually represents it's state. A CheckboxMenu can either
			have the valye "true" or "false", whereas a SelectMenu's value can take the form of any of it's children's 
			values.
		*/
		string getValue();

		/** \returns <string> the title of this Menu */
		string getTitle();

		/** set event handler 
			\param handler an ActionHandler object to take care of the events fired by this Menu
		*/
		void setActionHandler(ActionHandler* handler);

		/** recursively populates a ConfigFile object. If the name is nonempty, it will add the name,value tuple to the
			ConfigFile object. It then calls addValues on all it's children.
			\param config a ConfigFile object to add the values to.
		*/
		virtual void addValues(ConfigFile* config);

		/** reads values for this node and all children recursively from a ConfigFile object
			\param config a ConfigFile object to add the read the values from
		*/
		virtual void getValues(ConfigFile* config);

		/**
			Will stop the Menu from running by closing it and all it's children recursively.
		*/
		void close();

		/**
			\returns the parent of this Menu, or NULL if root
		*/
		Menu *getParent();

      virtual bool isSelectable();

		/**
			Sets wich item in the menu list is selected. Should be in range [0,number of child elements>
		*/
		void setSelected(int selected);

	protected:

		void setSurface(Surface *surface);
		virtual void draw();
		virtual void drawList(int x,int y,int w,int h);
		Surface* Menu::getSurface();

		typedef std::vector<Menu*> MenuList;

		enum {
			windowPadding = 8,
			listItemPadding = 4
		};

		int width,height,x,y;
		int selected;
		int scroll;
		Surface *surface;
		MenuList children;

		Menu *parent;

		bool running;

		Image *icon,*secondaryIcon;
		Image *background;

		string name,value,title;

		ActionHandler* actionHandler;
	};


};

#endif

